--[[

	Kui Lib
	By Kesava
	
	Just some helpful functions and tables.

]]

local addon, ns = ...

if kui then
	print('|c9900ff00Kui:|r an addon already possesses a global variable named kui.')
end

-- media files/functions unrelated to plugins
kui = {}

--- media / files ----------------------------------------------------------
local media = "Interface\\AddOns\\Kui\\media\\"
kui.m = {
	t = {
		-- borders
		square	= media .. 't\\simpleSquare',
		shadow	= media .. 't\\shadowBorder',
	
		-- textures
		solid		= media .. 't\\solid',
		innerShade	= media .. 't\\innerShade',
		
		-- progress bars
		bar		= media .. 't\\bar',
		sbar	= media .. 't\\barSmall',
	},
	f = {
		norm	= NAMEPLATE_FONT,
		accid	= media .. 'f\\accid.ttf',
		yanone	= media .. 'f\\yanone.ttf',
	},
}

--- helper functions -------------------------------------------------------

kui.GetClassColour = function(class, str)
	if not class then
		class = select(2, UnitClass('player'))
	elseif not RAID_CLASS_COLORS[class] then
		-- assume class is a unit
		class = select(2, UnitClass(class))
	end
	
	if CUSTOM_CLASS_COLORS then
		class = CUSTOM_CLASS_COLORS[class]
	else
		class = RAID_CLASS_COLORS[class]
	end
	
	if str then
		return string.format("%02x%02x%02x", class.r*255, class.g*255, class.b*255)
	else
		return class
	end
end

kui.UnitIsPet = function(unit)
	return (not UnitIsPlayer(unit) and UnitPlayerControlled(unit))
end

kui.GetUnitColour = function(unit, str)
	-- class colour for players or pets
	-- faction colour for NPCs
	local ret, r, g, b
	
	if (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit))
		or UnitIsDeadOrGhost(unit)
		or not UnitIsConnected(unit)
	then
		ret = { r = .5, g = .5, b = .5 }
	else
		if UnitIsPlayer(unit) or kui.UnitIsPet(unit) then
			return kui.GetClassColour(unit, str)
		else
			r, g, b = UnitSelectionColor(unit)
			ret = { r = r, g = g, b = b }
		end
	end
	
	if str then
		return string.format("%02x%02x%02x", ret.r*255, ret.g*255, ret.b*255)
	else
		return ret
	end
end

kui.UnitLevel = function(unit)
	level, cl	= UnitLevel(unit), UnitClassification(unit)
	l			= level
	
	if l < 0 then
		level	= '??'
		l		= 999
	end
	
	if cl == 'elite' then
		cl = '+'
	elseif cl == 'rareelite' then
		cl = 'r+'
	elseif cl == 'rare' then
		cl = 'r'
	elseif cl == 'worldboss' then
		cl = 'b'
	else
		cl = ''
	end
	
	return level, cl, GetQuestDifficultyColor(l)
end

kui.CreateFontString = function(parent, args)
	local ob, font, size, outline, alpha, shadow
	args = args or {}

	if args.reset then
		-- to change an already existing fontString
		ob = parent
	else
		ob = parent:CreateFontString(nil, 'OVERLAY')
	end

	font	= args.font or kui.m.f.norm
	size	= args.size or 12
	outline	= args.outline or nil
	alpha	= args.alpha or 1
	shadow	= args.shadow or false
	
	ob:SetFont(font, size, outline)
	ob:SetAlpha(alpha)
	
	if shadow then
		ob:SetShadowColor(0, 0, 0, 1)
		ob:SetShadowOffset(type(shadow) == 'table' and unpack(shadow) or 1, -1)
	elseif not shadow and args.reset then
		-- remove the shadow
		ob:SetShadowColor(0, 0, 0, 0)
	end
	
	return ob
end

-- Format numbers
kui.num = function(num)
	if num < 1000 then
		return num
	elseif num >= 1000000 then
		return string.format('%.1fm', num/1000000)
	elseif num >= 1000 then
		return string.format('%.1fk', num/1000)
	end
end

-- Format times (given in seconds)
kui.FormatTime = function(s)
	if s > 86400 then
		-- days
		return ceil(s/86400) .. 'd', s%86400
	elseif s >= 3600 then
		-- hours
		return ceil(s/3600) .. 'h', s%3600
	elseif s >= 60 then
		-- minutes
		return ceil(s/60) .. 'm', s%60
	elseif s <= 10 then
		return ceil(s), s - format("%.1f", s)
	end
	
	return floor(s), s - floor(s)
end

-- Pluralise a word pertaining to a value
kui.Pluralise = function(word, value, with)
	if value == 1 then
		return word
	else
		return word .. (with and with or 's')
	end
end

-- substr for utf8 characters (which are somtimes longer than 1 byte)
kui.utf8sub = function(string, fin)
	local b = string:len()
	
	if b <= fin then
		return string
	else
		local len, index = 0, 1
		
		while index <= b do
			len = len + 1
			
			local c = string:byte(index)
			if c > 240 then
				index = index + 4
			elseif c > 225 then
				index = index + 3
			elseif c > 192 then
				index = index + 2
			else
				index = index + 1
			end
			
			if len == fin then
				break
			end
		end
		
		if len == fin and index <= b then
			return string:sub(1, index - 1)
		else
			return string
		end
	end
end

string.kuisub = kui.utf8sub

--- frame ------------------------------------------------------------------
--- for event/script helpers
local kf = CreateFrame('Frame')

-- plugin functions container (used as metatable)
local kpf = {}
kpf.__index = kpf

function kpf:print(m)
	if type(m) == 'string' then
		print('|cff9900ff' .. (self.n or '[No name]') .. ':|r ' .. m)
	end
end

-- plugin helper ---------------------------------------------------------------
kui.plugins = {}

-- create metatable for a plugin
-- called like: example.kui = kui.addPlugin("My_Name", example)
function kui:addPlugin(name, table)
	if not name then
		kf.kui:print('attempted to add nil plugin')
		return
	end
		
	kui.plugins[name] = {}
	setmetatable(kui.plugins[name], kpf)
	
	kui.plugins[name].t = table
	kui.plugins[name].n = name
	
	return kui.plugins[name]
end

-- return a plugin's table
function kui:getPlugin(name)
	if not name then
		kf.kui:print('attempted to get nil plugin')
		return
	end

	if kui.plugins[name] then
		return kui.plugins[name].t
	else
		kf.kui:print('attempted to get non-existing plugin `' .. name .. '`')
		return false
	end
end

--- periodic tasks helper --------------------------------------------------
kui.tasks = {}

function kpf:addTask(interval)
	if not self.t.OnUpdate then
		kf.kui:print('no OnUpdate handler exists in plugin `' .. self.n .. '`')
		return
	end

	kui.tasks[self.n] = { i = interval, e = 0 }
end

function kpf:removeTask()
	kui.tasks[self.n] = nil
end

kf.elapsed = 0
kf:SetScript('OnUpdate', function(self, e)
	local p, t, pn
	self.elapsed = self.elapsed + e	
	
	if self.elapsed >= .1 then
		for pn, t in pairs(kui.tasks) do
			p = kui:getPlugin(pn)
		
			if p and p.OnUpdate then
				t.e = t.e + e
				
				if t.i == .1 or t.e >= t.i then
					-- (as .1 is the smallest possible, it needs to be synced)
					p:OnUpdate()
					t.e = 0
				end
			end
		end
		
		self.elapsed = 0
	end
end)

--- event handler helper ---------------------------------------------------
kui.events = {}

-- args: event[, function]
-- if a function is given it will be called instead of the named table value
-- it must accept the arguments: event, ...
function kpf:addEvent(e, f)
	if not f and not self.t[e] then
		kf.kui:print('no event handler exists for `' .. e .. '` in plugin `' .. self.n .. '`')
		return
	end

	if kui.events[e] then
		kui.events[e][self.n] = f or true
	else
		kui.events[e] = {
			[self.n] = f or true
		}
	end
	
	if not kf:IsEventRegistered(e) then
		kf:RegisterEvent(e)
	end
end

function kpf:removeEvent(e)
	if kui.events[e] then
		kui.events[e][self.n] = nil
		
		if #kui.events[e] == 0 then
			kf:UnregisterEvent(e)
		end
	end	
end

kf:SetScript('OnEvent', function(self, event, ...)
	if not kui.events[event] then return end

	for pn, enabled in pairs(kui.events[event]) do
		p = kui:getPlugin(pn)
	
		if (p[event] and enabled) or (type(enabled) == 'function') then
			if event ~= 'ADDON_LOADED' or (event == 'ADDON_LOADED' and ... == pn) then
				-- if event is addon_loaded, only run this if it is called
				-- for this particular addon (as always).
				if type(enabled) == 'function' then
					-- call the registered function
					enabled(event, ...)
				else
					p[event](p, ...)
				end
			end
		end
	end
end)

--- slash command helper ---------------------------------------------------
function kpf:addSlash(name, commands)
	if type(commands) == 'string' then
		commands = { commands }
	end

	if type(commands) ~= 'table' then
		kf.kui:print('invalid call to addSlash: commands must be table')
		return
	end

	if not self.t.Slash then
		kf.kui:print('no .Slash handler exists in plugin `' .. self.n .. '`')
		return
	end
	
	local i = 0
	for _, command in pairs(commands) do
		i = i + 1
		_G['SLASH_' .. name .. i] = '/' .. command
	end
	
	SlashCmdList[name] = function(msg)
		local wlist = {}
		for word in msg:gmatch('%S+') do
			table.insert(wlist, word)
		end
		self.t:Slash(msg, wlist)
	end
end

--[[ slash configuration helper ------------------------------------------------
	Registers the following variables in plugins:
	.saved and .savedPC
		These store the names of the global saved variables.
		
	.config and .configPC
		These store what will become a copy of the global saved variable once
		all settings have been added using the addSetting and addSettingGroup
		functions. This is commited to the global variable whenever a call to
		configSlash (the slash handler) is successful.  Storing this 'copy'
		enures that if a setting or group is renamed, it will not exist as a
		ghost in the global variable.
		
	
	TODO rather than using perChar, maybe just allow many variable names to be
	stored in .saved (a table), as the only purpose of perChar is so kui can
	differentiate the two; there is no logic to actually facilitate its
	named purpose. Maybe use an id or metatable to provide functions for this.

]]
-- configuration functions
kui.config = {}

-- generic argument functions
-- parse .1 .2 .3[ .4] into {.1, .2, .3[, .4] }
kui.config.ParseColourArg = function(self, msgs)
	local alpha = #self.v == 4
	local r, g, b, a = select(2, unpack(msgs))
	
	if alpha then
		a = tonumber(a)
	end
	
	r, g, b =
		tonumber(r),
		tonumber(g),
		tonumber(b)
	
	if	(r and r <= 1) and (g and g <= 1) and (b and b <= 1) and
		(alpha and (a and a <= 1) or not alpha)
	then
		self.v = { r, g, b, a or nil }
		return true
	end
end

-- convert a numeric string and validate it against provided limits
kui.config.ParseNumberArg = function(self, msgs)
	local num = tonumber(msgs[2])
	if	num and
		(not self.numL or num >= self.numL) and	-- lower limit
		(not self.numU or num <= self.numU)		-- upper limit
	then
		self.v = num
		return true
	end
end

-- invert the setting's boolean value
kui.config.ParseBooleanArg = function(self, msgs)
	self.v = not self.v
	return true
end

-- print the given generated configuration table
kui.config.PrintTable = function(tbl, altered)
	for setting, value in pairs(tbl) do
		if	type(value) == 'table' and
			(not altered or setting == altered)
		then
			local final -- final message line for this setting
			
			if value.gd then
				final = setting
			else
				final =
					(altered and setting..' is now' or setting)..
					(value.v == nil and '' or ': ')
			end
			
			if value.gd then
				final = final..': |cffffffaa'..value.gd..' settings|r'
			elseif value.v == true then
				final = final..'|cff00ff00enabled|r'
			elseif value.v == false then
				final = final..'|cffff0000disabled|r'
			elseif type(value.v) == 'table' then
				if #value.v == 3 or #value.v == 4 then
					-- value is/should be a colour table
					final = final..string.format('|cff%02x%02x%02x'..table.concat(value.v, ', ')..'|r',
						value.v[1]*255, value.v[2]*255, value.v[3]*255)
				else
					final = final..table.concat(value.v, ', ')
				end
			elseif value.v and (type(value.v) == 'string' or type(value.v) == 'number') then
				final = final..value.v
			end
			
			if value.d then
				final = final..' |cff888888['..value.d..']|r'
			end
			
			if value.rl then
				final = final..' |cffff0000[Requires UI reload]|r'
			end
			
			print(final)
		end
	end
end

-- register a saved variable
function kpf:registerSaved(variableName, perChar, version)
	if not _G[variableName] then
		_G[variableName] = {}
	end

	if version then
		if select('#', _G[variableName]) == 0 then
			_G[variableName].kcversion = version
		elseif	_G[variableName].kcversion ~= version then
			-- settings are out of date; reset them
			_G[variableName] = { kcversion = version }
			self:print((perChar and 'Per-character' or 'Global')..' configuration has been reset due to an update.')
		end
	end
	
	if perChar then
		self.savedPC = variableName
	else
		self.saved = variableName
	end
end

-- update the registered saved variable
function kpf:updateSaved(perChar)
	local var = perChar and self.savedPC or self.saved

	if var then
		_G[var] = perChar and kui.configPC or kui.config
	end
end

-- add a setting group to this plugin's config table
function kpf:addSettingGroup(name, description, perChar)
	if not self.saved and not self.savedPC then
		kf.kui:print('attempt to add setting group with no saved variable in plugin `'..self.n..'`')
		return
	end
	
	if name == 'kcversion' then
		kf.kui:print('attempt to add a setting group named `kcversion` in plugin `'..self.n..'`')
		return
	end
	
	local tbl
	if perChar then
		self.configPC = self.configPC or {}
		tbl = self.configPC
	else
		self.config = self.config or {}
		tbl = self.config
	end
	
	-- create group
	tbl[name] = { gd = description }
end

-- add a setting to this plugin's config table
function kpf:addSetting(group, command, default, desc, func, add)
	if not self.saved and not self.savedPC then
		kf.kui:print('attempt to add setting with no saved variable in plugin `'..self.n..'`')
		return
	end

	--[[ add = {
		i = invalidMsg
		p = perChar
		l = loadFunc
	}]]
	
	local perChar	= add and add.p or nil
	local sv		= perChar and _G[self.savedPC] or _G[self.saved]
	
	local tbl
	if perChar then
		self.configPC = self.configPC or {}
		tbl = self.configPC
	else
		self.config = self.config or {}
		tbl = self.config
	end
	
	if group then
		if type(tbl[group]) == 'table' and tbl[group].gd then
			-- create setting in the group
			tbl = tbl[group]
			sv = sv[group] or nil
		else
			kf.kui:print('attempt to add setting to non-existing or corrupt group `'..group..'` in plugin `'..self.n..'`')
			return
		end
	end

	-- create command table
	tbl[command] = {
		d = desc,
		f = func or kui.config.ParseBooleanArg
	}
	tbl = tbl[command]
	
	if sv and sv[command] then
		-- this setting exists in the saved variable, inherit it
		tbl.v = sv[command].v
	else
		-- or use the default value
		tbl.v = default
	end
	
	if add then
		-- add additional values
		for k,v in pairs(add) do
			tbl[k] = v
		end
		
		if tbl.l then
			-- run load function
			tbl.l(tbl)
		end
	end
end

-- handle slash command for configuration (print, call functions, etc)
-- TODO separate global from perchar by using two PrintTables
function kpf:configSlash(msg, msgs)
	local tbl, altered = self.config
	
	if _G[self.saved].kcversion and not self.config.kcversion then
		-- store the configuration version in the local copy
		self.config.kcversion = _G[self.saved].kcversion
	end
	
	if type(tbl[msgs[1]]) == 'table' and tbl[msgs[1]].gd then
		-- this is a call to a grouped setting
		tbl = tbl[msgs[1]]
		table.remove(msgs, 1)
	end
	
	for setting, value in pairs(tbl) do
		if	msgs[1] ~= nil and 
			type(value) == 'table' and
			msgs[1] == setting and
			value.f
		then
			altered = value.f(value, msgs)
			
			if altered then
				altered = setting
			elseif altered == nil then
				self.t.kui:print('|cffff0000Invalid arguments for '..setting..'.|r')
				
				if value.i and type(value.i) == 'string' then
					print('|cffff0000Usage:|r '..value.i)
				end
				return
			end
			
			if value.l then
				-- run load function
				value.l(value)
			end
			
			break
		end
	end
	
	-- print config
	self.t.kui:print('configuration:')
	kui.config.PrintTable(tbl, altered)
	
	-- update saved variable
	_G[self.saved] = self.config
end

function kpf:getSaved(group, setting)
	if self.config[group] and self.config[group][setting] then
		return self.config[group][setting].v
	elseif self.config[setting] then
		return self.config[setting].v
	else
		return nil
	end
end

--- finishing up -----------------------------------------------------------

kf.Slash = function(msg, editbox)
	ReloadUI()
end

--- now register it as a plugin of itself? ---------------------------------

kf.kui = kui:addPlugin(addon, kf)
kf.kui:addSlash('RELOADUI', 'rl')